import torch
from d2l import torch as d2l
from torch import nn
from model.decomposable_attention import DecomposableAttention

class Train():
    def __init__(self):
       pass
      
    def run_training(self, train_iter, test_iter,vocab, learning_rate, 
                            epochs, embed_size, num_hiddens,device,
                            save_result: bool = True,
                            save_path:str = "trained_model/model.pth",
                            vocab_path:str = "trained_model/vocab.pth"):         
        net = DecomposableAttention(vocab, embed_size, num_hiddens)
        glove_embedding = d2l.TokenEmbedding('glove.6b.100d')
        embeds = glove_embedding[vocab.idx_to_token]
        net.embedding.weight.data.copy_(embeds)
        trainer = torch.optim.Adam(net.parameters(), lr=learning_rate)
        loss = nn.CrossEntropyLoss(reduction="none")   
        d2l.train_ch13(net, train_iter, test_iter, loss, trainer, epochs, device)
        
        if save_result:
          torch.save(net.state_dict(), save_path)
          torch.save(vocab, vocab_path)



import torch
from d2l import torch as d2l
from torch import nn
from model.decomposable_attention import DecomposableAttention
from sklearn.metrics import classification_report

class Predict():
    def __init__(self, embed_size, num_hiddens,
                  model_path="trained_model/model.pth",
                  vocab_path="trained_model/vocab.pth"):
      self.embed_size = embed_size
      self.num_hiddens = num_hiddens
      self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
      self.vocab = torch.load(vocab_path)
      self.model = DecomposableAttention(self.vocab, self.embed_size,self.num_hiddens)
      self.model.load_state_dict(torch.load(model_path)) 
      self.model.to(self.device)     
      self.model.eval()
      
    def make_single_prediction(self, premise, hypothesis):         
      premise = torch.tensor(self.vocab[premise], device=d2l.try_gpu())
      hypothesis = torch.tensor(self.vocab[hypothesis], device=d2l.try_gpu())
      label = torch.argmax(self.model([premise.reshape((1, -1)),
                           hypothesis.reshape((1, -1))]), dim=1)
      return 'Entailment' if label == 0 else 'Contradiction' if label == 1 \
            else 'neutral'

    def make_multiple_prediction(self, data):
      y_hat=[]
      for i in range(len(data)):
          y_hat.append(self.make_single_prediction(data.iloc[i, :]["premise"].split(), \
                                data.iloc[i, :]["hypotheis"].split()))
      
      return y_hat
    
    def make_multiple_prediction_with_classification_report(self, data):
      def change_label(x):
        if x == "Entailment":
          return 0
        elif x =="Contradiction":
          return 1
        else:
          return 2
      y_hat =self.make_multiple_prediction(data);
      y_hat_transformed = [change_label(x) for x in y_hat]
      labels = data["label"].values
      print(f"\n {classification_report(labels, y_hat_transformed, labels=[0,1,2])}")


    



from yaml.loader import SafeLoader
import yaml
from d2l import torch as d2l

from preprocessing.data_management import DataService
from train import Train
from predict import Predict

# Open the file and load the file
with open('config/config.yaml') as f:
    config = yaml.load(f, Loader=SafeLoader)
    print(config)

args = config['Train']
ds = DataService()
train_set,test_set= ds.load_data('data/train.csv', 'data/test.csv')
device = d2l.try_all_gpus()
num_workers = d2l.get_dataloader_workers()
train_iter,test_iter,vocab= ds.create_snli_dataset(train=train_set, \
                                                   test=test_set, num_workers=num_workers)

#Train
tr = Train()
tr.run_training(train_iter, test_iter, vocab,args["learning_rate"],\
                args["epochs"],args["embed_size"],args["num_hiddens"], device) 


#Predict
pr = Predict(args["embed_size"],args["num_hiddens"])
pr.make_multiple_prediction_with_classification_report(test_set)

#Make Single Prediction
test_set.iloc[4, :]
pr.make_single_prediction(test_set.iloc[4, :]["premise"].split(), \
             test_set.iloc[4, :]["hypotheis"].split())


import pandas as pd
import torch
import d2l
from .snli_dataset import SNLIDataset

class DataService():
    def __init__(self): 
       pass 

    #loads csv into pandas dataframe.
    def load_data(self, train_path, test_path):    
        train = pd.read_csv(train_path)
        test = pd.read_csv(test_path)
        train["label"] = train["label"].astype(int)
        test["label"] = test["label"].astype(int)
        return train, test

    #load pandas dataframe to dataset.
    def create_snli_dataset(self, train, test, num_steps = 50, batch_size = 256, num_workers = 4):        
        train_set = SNLIDataset(train, num_steps)
        test_set = SNLIDataset(test, num_steps, train_set.vocab)
        vocab = train_set.vocab
        train_iter = torch.utils.data.DataLoader(train_set, batch_size,
                                                    shuffle=False,
                                                    num_workers=num_workers)
        test_iter = torch.utils.data.DataLoader(test_set, batch_size,
                                                shuffle=False,
                                                num_workers=num_workers)
        return train_iter, test_iter, vocab


import torch
from model.mlp import mlp
from torch import nn
from torch.nn import functional as F

class Aggregate(nn.Module):
    def __init__(self, num_inputs, num_hiddens, num_outputs, **kwargs):
        torch.manual_seed(123)
        super(Aggregate, self).__init__(**kwargs)
        self.h = mlp(num_inputs, num_hiddens, flatten=True)
        self.linear = nn.Linear(num_hiddens, num_outputs)

    def forward(self, V_A, V_B):
        torch.manual_seed(123)
        # Sum up both sets of comparison vectors
        V_A = V_A.sum(dim=1)
        V_B = V_B.sum(dim=1)
        # Feed the concatenation of both summarization results into an MLP
        Y_hat = self.linear(self.h(torch.cat([V_A, V_B], dim=1)))
        return Y_hat


import torch
from torch import nn
from .attend import Attend
from .compare import Compare
from .aggregate import Aggregate
from torch.nn import functional as F

"""
Putting all the building blocks together in the following module
"""
class DecomposableAttention(nn.Module):
    def __init__(self, vocab, embed_size, num_hiddens, num_inputs_attend=100,
                 num_inputs_compare=200, num_inputs_agg=400, **kwargs):
        torch.manual_seed(123)
        super(DecomposableAttention, self).__init__(**kwargs)
        self.embedding = nn.Embedding(len(vocab), embed_size)
        self.attend = Attend(num_inputs_attend, num_hiddens)
        self.compare = Compare(num_inputs_compare, num_hiddens)
        # There are 3 possible outputs: entailment, contradiction, and neutral
        self.aggregate = Aggregate(num_inputs_agg, num_hiddens, num_outputs=3)

    def forward(self, X):
        torch.manual_seed(123)
        premises, hypotheses = X
        A = self.embedding(premises)
        B = self.embedding(hypotheses)
        beta, alpha = self.attend(A, B)
        V_A, V_B = self.compare(A, B, beta, alpha)
        Y_hat = self.aggregate(V_A, V_B)
        return Y_hat


import torch
from d2l import torch as d2l

''''
This class is responsible for creating SNLI dataset from pandas dataframe.
'''
class SNLIDataset(torch.utils.data.Dataset):
    """A customized dataset to load the SNLI dataset."""
    def __init__(self, dataset, num_steps, vocab=None):
        self.num_steps = num_steps
        all_premise_tokens = d2l.tokenize(dataset["hypotheis"])
        all_hypothesis_tokens = d2l.tokenize(dataset["premise"])
        if vocab is None:
            self.vocab = d2l.Vocab(all_premise_tokens + all_hypothesis_tokens,
                                   min_freq=5, reserved_tokens=['<pad>'])
        else:
            self.vocab = vocab
        self.premises = self._pad(all_premise_tokens)
        self.hypotheses = self._pad(all_hypothesis_tokens)
        self.labels = torch.tensor(dataset["label"])
        print('read ' + str(len(self.premises)) + ' examples')

    def _pad(self, lines):
        torch.manual_seed(123)
        return torch.tensor([d2l.truncate_pad(
            self.vocab[line], self.num_steps, self.vocab['<pad>'])
                         for line in lines])

    def __getitem__(self, idx):
        torch.manual_seed(123)
        return (self.premises[idx], self.hypotheses[idx]), self.labels[idx]

    def __len__(self):
        torch.manual_seed(123)
        return len(self.premises)



import torch
from model.mlp import mlp
from torch import nn
from torch.nn import functional as F

class Compare(nn.Module):
    def __init__(self, num_inputs, num_hiddens, **kwargs):
        torch.manual_seed(123)
        super(Compare, self).__init__(**kwargs)
        self.g = mlp(num_inputs, num_hiddens, flatten=False)

    def forward(self, A, B, beta, alpha):
        torch.manual_seed(123)
        V_A = self.g(torch.cat([A, beta], dim=2))
        V_B = self.g(torch.cat([B, alpha], dim=2))
        return V_A, V_B



import torch
from torch import nn
from torch.nn import functional as F

def mlp(num_inputs, num_hiddens, flatten):
    torch.manual_seed(123)
    net = []
    net.append(nn.Dropout(0.2))
    net.append(nn.Linear(num_inputs, num_hiddens))
    net.append(nn.ReLU())
    if flatten:
        net.append(nn.Flatten(start_dim=1))
    net.append(nn.Dropout(0.2))
    net.append(nn.Linear(num_hiddens, num_hiddens))
    net.append(nn.ReLU())
    if flatten:
        net.append(nn.Flatten(start_dim=1))
    return nn.Sequential(*net)


import torch
from torch import nn
from torch.nn import functional as F
from model.mlp import mlp

class Attend(nn.Module):
    def __init__(self, num_inputs, num_hiddens, **kwargs):
        torch.manual_seed(123)
        super(Attend, self).__init__(**kwargs)
        self.f = mlp(num_inputs, num_hiddens, flatten=False)

    def forward(self, A, B):
        torch.manual_seed(123)
        # Shape of `A`/`B`: (`batch_size`, no. of tokens in sequence A/B,
        # `embed_size`)
        # Shape of `f_A`/`f_B`: (`batch_size`, no. of tokens in sequence A/B,
        # `num_hiddens`)
        f_A = self.f(A)
        f_B = self.f(B)
        # Shape of `e`: (`batch_size`, no. of tokens in sequence A,
        # no. of tokens in sequence B)
        e = torch.bmm(f_A, f_B.permute(0, 2, 1))
        # Shape of `beta`: (`batch_size`, no. of tokens in sequence A,
        # `embed_size`), where sequence B is softly aligned with each token
        # (axis 1 of `beta`) in sequence A
        beta = torch.bmm(F.softmax(e, dim=-1), B)
        # Shape of `alpha`: (`batch_size`, no. of tokens in sequence B,
        # `embed_size`), where sequence A is softly aligned with each token
        # (axis 1 of `alpha`) in sequence B
        alpha = torch.bmm(F.softmax(e.permute(0, 2, 1), dim=-1), A)
        return beta, alpha